import pathlib

from .decorator import *

from flask import Blueprint

panel_package_path = pathlib.Path(__file__).parent

mod = Blueprint(
	'panel',
	'panel',
	static_folder=str(panel_package_path / 'static'),
	template_folder=str(panel_package_path / 'templates'),
	url_prefix='/panel'
)


import math

from sqlalchemy import or_

from flask import (request, render_template, redirect, url_for,
                   jsonify, abort, current_app, flash)
from flask_login import current_user, login_required
from flask_babel import gettext, pgettext

from fardel.core.auth.models import User, Group, Permission
from fardel.ext import db

from .. import mod, staff_required, admin_required, permission_required


#########
# USERS #
#########

@mod.route('/auth/users/search/email/')
@permission_required("can_send_email")
@staff_required
@login_required
def email_users_search():
    like_query = request.args.get("q")
    users = User.query.filter(User._email.like("{}%".format(like_query))).all()
    return jsonify({
        "users": [{"text": u.email} for u in users],
    })


@mod.route('/auth/users/list/')
@permission_required("can_get_users")
@staff_required
@login_required
def users_list():
    page = request.args.get('page', type=int, default=1)
    per_page = request.args.get('per_page', type=int, default=40)
    query = User.query.filter_by(is_staff=False, is_admin=False)
    pages = math.ceil(query.count() / per_page)
    users = query.paginate(
        page=page, per_page=per_page, error_out=False).items
    return render_template('auth/users_list.html',
                           users=users, pages=pages, page=page)


@mod.route('/auth/users/edit/<int:user_id>/', methods=['POST', 'GET'])
@admin_required
@staff_required
@login_required
def users_edit(user_id):
    user = User.query.filter_by(id=user_id).first_or_404()
    if request.method == "POST":
        email = request.form.get('email')
        first_name = request.form.get('first_name')
        last_name = request.form.get('last_name')
        password = request.form.get('password')
        confirmed = request.form.get('confirmed', type=bool)
        is_admin = request.form.get('is_admin', type=bool)
        is_staff = request.form.get('is_staff', type=bool)
        group_ids = request.form.getlist('group_ids')

        if not email or not first_name or not last_name:
            flash('gettext(email, first name, last name and password fields can not be empty!)', 'error')
        user._email = email
        user.first_name = first_name
        user.last_name = last_name
        user.groups = []
        for gid in group_ids:
            group = Group.query.filter_by(id=gid).first()
            user.groups.append(group)

        if password and password != "":
            user.password = password

        user.confirmed = confirmed
        user.is_admin = is_admin
        user.is_staff = is_staff
        db.session.commit()

        if user.is_staff or user.is_admin:
            return redirect(url_for("panel.staffs_list"))
        else:
            return redirect(url_for("panel.users_list"))

    groups = Group.query.all()
    user_group_ids = [g.id for g in user.groups]
    return render_template('auth/users_form.html', user=user, groups=groups, user_group_ids=user_group_ids)


@mod.route('/auth/users/create/', methods=['POST', 'GET'])
@admin_required
@staff_required
@login_required
def users_create():
    if request.method == "POST":
        email = request.form.get('email')
        if User.query.filter_by(_email=email).first():
            flash(gettext('user with this email already exists.'), 'error')
            return redirect(url_for('panel.users_create'))

        first_name = request.form.get('first_name')
        last_name = request.form.get('last_name')
        password = request.form.get('password')
        confirmed = request.form.get('confirmed', type=bool)
        is_admin = request.form.get('is_admin', type=bool)
        group_ids = request.form.getlist('group_ids')

        for gid in group_ids:
            group = Group.query.filter_by(id=gid).first()
            user.groups.append(group)

        if not email or not first_name or not last_name:
            flash(gettext('email, first name, last name and password fields can not be empty!'), 'error')
            return redirect(url_for('panel.users_create'))

        user = User(_email=email, first_name=first_name, last_name=last_name, password=password,
                    confirmed=confirmed, is_admin=is_admin, group_id=group_id)

        db.session.add(user)
        db.session.commit()
        flash(gettext('user successfully created'), 'success')
        if user.is_staff or user.is_admin:
            return redirect(url_for("panel.staffs_list"))
        else:
            return redirect(url_for("panel.users_list"))
    groups = Group.query.all()
    return render_template('auth/users_form.html', groups=groups)


@mod.route('/auth/users/delete/<int:user_id>/')
@admin_required
@staff_required
@login_required
def users_delete(user_id):
    abort(404)

################
# STAFFMEMBERS #
################


@mod.route('/auth/staffs/list/')
@permission_required("can_get_users")
@admin_required
@staff_required
@login_required
def staffs_list():
    page = request.args.get('page', type=int, default=1)
    per_page = request.args.get('per_page', type=int, default=40)
    query = User.query.filter(or_(User.is_staff == True, User.is_admin == True))
    pages = math.ceil(query.count() / per_page)
    users = query.paginate(
        page=page, per_page=per_page, error_out=False).items
    return render_template('auth/users_list.html',
                           users=users, pages=pages, page=page)


###############
# GROUPS #
###############


@mod.route('/auth/permissions/list/')
@permission_required('can_get_permissions')
@staff_required
@login_required
def permissions_get():
    return jsonify({
        'permissions': [obj.dict() for obj in
                        Permission.query.all()]
    })

###########
# GRPOUPS #
###########


@mod.route('/auth/groups/list/')
@permission_required('can_get_groups')
@staff_required
@login_required
def groups_list():
    page = request.args.get('page', type=int, default=1)
    per_page = request.args.get('per_page', type=int, default=40)
    query = Group.query
    pages = math.ceil(query.count() / per_page)
    groups = query.paginate(
        page=page, per_page=per_page, error_out=False).items
    return render_template('auth/groups_list.html',
                           groups=groups, pages=pages, page=page)


@mod.route('/auth/group/create/', methods=["POST", "GET"])
@admin_required
@staff_required
@login_required
def groups_create():
    permissions = Permission.query.all()
    if request.method == "POST":
        name = request.form.get("name")
        permissions = request.form.getlist("permissions")

        group = Group(name=name)

        for pid in permissions:
            permission = Permission.query.filter_by(id=pid).first()
            group.permissions.append(permission)

        db.session.add(group)
        db.session.commit()
        flash(gettext("Group created successfully"))
        return redirect(url_for("panel.groups_list"))

    return render_template("auth/groups_form.html", permissions=permissions)


@mod.route('/auth/group/edit/<int:group_id>/', methods=["POST", "GET"])
@admin_required
@staff_required
@login_required
def groups_edit(group_id):
    group = Group.query.filter_by(id=group_id).first_or_404()
    if request.method == "POST":
        name = request.form.get("name")
        permissions = request.form.getlist("permissions")

        group.name = name
        group.permissions = []
        for pid in permissions:
            permission = Permission.query.filter_by(id=pid).first()
            group.permissions.append(permission)

        db.session.add(group)
        db.session.commit()
        flash(gettext("Group editted successfully"))
        return redirect(url_for("panel.groups_list"))

    permissions = Permission.query.all()
    group_permission_ids = [p.id for p in group.permissions]
    return render_template("auth/groups_form.html", permissions=permissions, group=group, group_permission_ids=group_permission_ids)



from flask import redirect, url_for, render_template, abort, request, flash
from flask_login import (current_user, login_user,
                         logout_user, login_required, logout_user)
from flask_babel import gettext

from fardel.core.auth.models import User

from .. import mod

from . import auth
from . import media
from . import email

from ..sidebar import panel_sidebar, Section, Link, ChildLink


@mod.before_app_first_request
def add_home_section():
    section = Section('')
    section.add_link(Link('fas fa-home', gettext('home'), url_for('panel.home')))
    panel_sidebar.add_section(section)


@mod.before_app_first_request
def add_auth_section():
    section = Section(gettext("Authentication"))
    link = Link("fas fa-users", gettext("Users"))
    link.add_child(ChildLink(gettext("Users"), url_for('panel.users_list')))
    link.add_child(ChildLink(gettext("Staffmembers"), url_for('panel.staffs_list')))
    section.add_link(link)

    link = Link("fas fa-layer-group", gettext("Groups"), url_for('panel.groups_list'))
    section.add_link(link)
    panel_sidebar.add_section(section)

    section = Section(gettext("Communication"))
    link = Link("fas fa-envelope", gettext("Email"))
    link.add_child(ChildLink(gettext("Send Email"), url_for('panel.email_send')))
    section.add_link(link)

    panel_sidebar.add_section(section)


# @mod.before_app_first_request
# def add_media_section():
# 	section = Section(gettext("Uploads"))
# 	link = Link("fa fa-file", gettext("Files"))
# 	link.add_child(ChildLink(gettext("Files"), "#"))
# 	# link.add_child(ChildLink("آلبوم عکس", "#"))
# 	section.add_link(link)

# 	panel_sidebar.add_section(section)


@mod.route('/login/', methods=['POST', 'GET'])
def login():
    if current_user.is_authenticated:
        return redirect(url_for('panel.home'))

    if request.method == "POST":
        email = request.form.get('email')
        password = request.form.get('password')

        u = User.query.filter_by(_email=email).first()
        if u and (u.is_staff or u.is_admin) and u.check_password(password):
            login_user(u)
            return redirect(url_for('panel.home'))
        flash(gettext('Email or password is wrong...'))
    return render_template('login.html')


@mod.route('/logout/')
@login_required
def logout():
    logout_user()
    return redirect('/')


@mod.route('/')
def index():
    if not current_user.is_authenticated:
        abort(404)
    return redirect(url_for('panel.home'))


@mod.route('/home/')
@login_required
def home():
    return render_template('dashboard.html')



from flask import (request, render_template, redirect, url_for,
                   jsonify, abort, current_app, flash)
from flask_login import current_user, login_required
from flask_babel import gettext, pgettext

from fardel.core import email as email_pkg

from .. import mod, staff_required, permission_required


@mod.route('/communication/email/send/', methods=["GET", "POST"])
@permission_required("can_send_email")
@staff_required
@login_required
def email_send():
    if request.method == "POST":
        sender = request.form.get("sender")
        recievers = request.form.getlist("recievers")
        title = request.form.get("title")
        content = request.form.get("content")

        if not sanity_check_email_addresses(sender, *recievers):
            flash("Emails are not correct")
            return redirect(url_for("panel.email_send"))

        message = {
            "subject": title,
            "text": content,
            "html": content,
            "from": sender,
            "to": recievers,
        }
        email_pkg.send_email(message)

        flash("Messages are sent")
        return redirect(url_for("panel.email_send"))
    return render_template("communication/send_email.html")


def sanity_check_email_addresses(*emails):
    print(emails)
    for email in emails:
        if not "@" in email:
            return False
    return True



from flask_restful import Api, abort, Resource
from flask_babel import gettext

__all__ = (
	'create_api', 'abort'
)

errors = {
    'NoAuthorizationError': {
        'message': gettext("Missing Authorization header"),
        'status': 403,
    },
    'ExpiredSignatureError': {
    	'message': gettext("Token is already expired"),
    	'status': 401
    }
}

def create_api(module):
	return Api(module, errors=errors)


import logging

from flask import current_app
from flask_mail import Message

from fardel.ext import mail

from threading import Thread


logger = logging.getLogger(__file__)


def send_async_email(app, msg):
    try:
        with app.app_context():
            mail.send(msg)
    except Exception as err:
        logger.exception("Couldn't send email because of err: %s" % err, exc_info=True)


def send_email(email):
    app = current_app._get_current_object()
    msg = Message(
        subject=email['subject'],
        body=email['text'],
        html=email['html'],
        sender=email['from'],
        recipients=email['to'],
    )
    thr = Thread(target=send_async_email, args=[app, msg])
    thr.start()



from flask_login import current_user


class Section():
    def __init__(self, title, permission=None):
        self.title = title
        self.links = []

    def add_link(self, link):
        self.links.append(link)

    def get_link(self, title):
        for link in self.links:
            if link.title == title:
                return link

    @property
    def permissions(self):
        return [link.permission for link in self.links]

    def render_links(self):
        return "".join(
            [link.render() for link in self.links
             if current_user.can(link.permission)]
        )

    def render(self):
        return """<div class="menu_section">
<h3>{title}</h3>
<ul class="nav side-menu">
{links}
</ul>
</div>""".format(title=self.title, links=self.render_links())


class Link():
    def __init__(self, icon, title, href="#", permission=None):
        self.children = []
        self.title = title
        self.href = href
        self.icon = icon
        self.permission = permission

    def add_child(self, child):
        self.children.append(child)

    def fa_chevron_down(self):
        if self.children:
            return '<span class="fas fa-chevron-down"></span>'
        return ''

    def render_children(self):
        return "".join([child.render() for child in self.children
                        if current_user.can(child.permission)])

    def render(self):
        return """<li><a href="{href}"><i class="{icon}"></i> {title} {chevron}</a>
<ul class="nav child_menu">
{children}
</ul>
</li>""".format(icon=self.icon, title=self.title, href=self.href,
                children=self.render_children(), chevron=self.fa_chevron_down())


class ChildLink():
    def __init__(self, name, href, permission=None):
        self.name = name
        self.href = href
        self.permission = permission

    def render(self):
        return '<li><a href="{href}">{name}</a></li>'.format(
            href=self.href,
            name=self.name
        )


class PanelSidebar():
    def __init__(self):
        self.sections = []

    def add_section(self, section):
        self.sections.append(section)

    def get_section(self, title):
        for section in self.sections:
            if section.title == title:
                return section

    def render(self):
        sections = []
        for section in self.sections:
            found = False
            for perm in section.permissions:
                if current_user.can(perm):
                    sections.append(section)
                    break

        return "".join([section.render() for section in sections])


panel_sidebar = PanelSidebar()



"""
Where a RatSnake app is created.
"""

import logging
import json

import sqlalchemy
from raven.contrib.flask import Sentry

from flask import Flask, request, jsonify, redirect, url_for, render_template

from fardel import core
from fardel.ext import db, jwt, cache, login_manager, babel, mail


__all__ = (
    'Fardel',
)

DEFAULT_APP_NAME = 'fardel'


class Fardel(object):
    ignored_panel_urls = ['static']

    def __init__(self, config_class, **flask_options):
        self.app = Flask(DEFAULT_APP_NAME, **flask_options)
        self.panels = []

        self.configure_app(config_class)
        self.configure_addons()
        self.configure_views()
        self.configure_extentions()
        self.configure_logger()
        self.init_jinja_globals()
        self.configure_sentry()

    def configure_app(self, config_class):
        self.app.config.from_object(config_class)

    def configure_addons(self):
        from fardel.core.auth.views import mod as auth
        from fardel.core.panel.views import mod as panel
        from fardel.core.media.views import mod as media
        self.app.register_blueprint(auth)
        self.app.register_blueprint(panel)
        self.app.register_blueprint(media)

        fardel_apps_path = self.app.config['FARDEL_APP_PATH']

        for app_name in self.app.config['ACTIVE_APPS']:
            try:
                bp = __import__('{fardel_apps_path}.{app_name}'.format(
                    app_name=app_name,
                    fardel_apps_path=fardel_apps_path
                ), fromlist=['views', 'panel'])

                if hasattr(bp, 'panel'):
                    self._register_panel(self.app, bp.panel.mod, app_name)
                else:
                    self.app.logger.debug("No admin panel for %s" % app_name)

                try:
                    self.app.register_blueprint(bp.views.mod)
                    self.app.logger.debug("Module %s registered" % app_name)
                except Exception as err:
                    self.app.logger.warning("%s app doesn't have blueprint." % app_name)
                    self.app.logger.warning(err, exc_info=True)

            except Exception as err:
                self.app.logger.warning("%s app not found" % app_name)
                self.app.logger.warning(err, exc_info=True)

    def configure_views(self):
        self.app.add_url_rule('/api/search/', 'search', core.search)

    def configure_logger(self):
        self.app.logger = logging.getLogger("fardel")

    def configure_extentions(self):
        db.init_app(self.app)
        with self.app.app_context():
            db.configure_mappers()
        jwt.init_app(self.app)
        cache.init_app(self.app)
        login_manager.init_app(self.app)
        babel.init_app(self.app)
        mail.init_app(self.app)

    def configure_sentry(self):
        if self.app.config['SENTRY_DSN']:
            sentry = Sentry(self.app, dsn=self.app.config['SENTRY_DSN'])

    def init_jinja_globals(self):
        from fardel.core.panel.template_tags import add_globals
        add_globals(self.app)

    def _register_panel(self, app, blueprint, app_name):
        if blueprint.name.endswith("panel"):
            self.panels.append(blueprint.name)
            self.app.register_blueprint(blueprint)
            for rule in self.app.url_map.iter_rules():
                endpoint = rule.endpoint.split('.')
                if endpoint[0] in self.panels and not rule.rule.startswith("/panel/"):
                    raise Exception("Panel url must start with /panel/")
                if endpoint[0] in self.panels and endpoint[1] not in self.ignored_panel_urls:
                    view_func = self.app.view_functions.get(rule.endpoint)
                    has_staff_required = False
                    has_admin_required = False
                    if hasattr(view_func, 'decorators'):
                        has_staff_required = view_func.decorators.get('staff_required', False)
                        has_admin_required = view_func.decorators.get('admin_required', False)

                    if not has_staff_required and not has_admin_required:
                        raise Exception("No staff_required or admin_required for panel function %s" % rule.endpoint)

            self.app.logger.debug("Admin panel for %s registered" % app_name)
        else:
            raise Exception("Panel blueprint's name must ends with panel")



import pathlib

from flask import Blueprint

auth_package_path = pathlib.Path(__file__).parent

mod = Blueprint(
    'auth',
    'auth',
    static_folder=str(auth_package_path / 'static'),
    template_folder=str(auth_package_path / 'templates'),
    url_prefix="/api/auth",
)



from functools import wraps
from flask import abort

from flask_jwt_extended import current_user
from flask_login import current_user as fl_current_user
from flask_restful import abort as rest_abort

__all__ = [
    'permission_required',
    'staff_required',
    'staff_required_rest',
    'admin_required',
    'admin_required_rest',
]


def permission_required(permission):
    def decorator(f):
        @wraps(f)
        def decorated_function(*args, **kwargs):
            if fl_current_user and fl_current_user.can(permission):
                return f(*args, **kwargs)
            return abort(403)
        return decorated_function
    return decorator


def permission_required_rest(permission):
    def decorator(f):
        @wraps(f)
        def decorated_function(*args, **kwargs):
            if current_user and current_user.can(permission):
                return f(*args, **kwargs)
            return rest_abort(403)
        return decorated_function
    return decorator


def staff_required(func):
    if not hasattr(func, 'decorators'):
        func.decorators = {}
    func.decorators['staff_required'] = True

    @wraps(func)
    def wrapper(*args, **kwargs):
        if fl_current_user and (fl_current_user.is_staff or fl_current_user.is_admin):
            return func(*args, **kwargs)
        return abort(403)
    return wrapper


def staff_required_rest(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        if current_user and (current_user.is_staff or current_user.is_admin):
            return func(*args, **kwargs)
        return rest_abort(403)
    return wrapper


def admin_required(func):
    if not hasattr(func, 'decorators'):
        func.decorators = {}
    func.decorators['admin_required'] = True

    @wraps(func)
    def wrapper(*args, **kwargs):
        if fl_current_user and fl_current_user.is_authenticated and fl_current_user.is_admin:
            return func(*args, **kwargs)
        return abort(403)
    return wrapper


def admin_required_rest(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        if current_user and current_user.is_admin:
            return func(*args, **kwargs)
        return rest_abort(403)
    return wrapper



import time
import string
import random

from flask import request


def convert_timestamp(dt):
    return time.mktime(dt.timetuple())


def make_cache_key(*args, **kwargs):
    path = request.path
    args = str(hash(frozenset(request.args.items())))
    # lang = get_locale()
    return (path + args).encode('utf-8')


def cache_get_key(resource_cls):
    """ Decorator for adding resources to Api App """
    resource_cls.get.make_cache_key = make_cache_key
    return resource_cls


def random_string(length=6):
    """Generate a random string of fixed length """
    letters = string.ascii_uppercase
    return ''.join(random.choice(letters) for i in range(length))



import json
import time
import bcrypt

from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm.attributes import flag_modified

from flask_babel import gettext
from flask import current_app, jsonify, render_template
from flask_login import UserMixin

from fardel.ext import db, jwt, login_manager
from fardel.core.utils import random_string

__all__ = ['User', 'Permission', 'Group', 'RevokedToken', 'setup_permissions']

CONFIRM_EMAIL = "email_confirm"

TEMPLATES = {
    "confirm_email": {
        "txt": "confirm_email.txt",
        "html": "confirm_email.html",
    }
}


def setup_permissions():
    Group.setup_permissions()
    User.setup_permissions()
    Permission.setup_permissions()


class AbstractModelWithPermission():
    @classmethod
    def setup_permissions(cls):
        for permission in cls.Meta.permissions:
            p = Permission.query.filter_by(code_name=permission[0]).first()
            if not p:
                p = Permission(code_name=permission[0], name=permission[1])
                db.session.add(p)
                db.session.commit()


class Permission(db.Model, AbstractModelWithPermission):
    __tablename__ = "auth_permissions"
    id = db.Column(db.Integer, primary_key=True, index=True)
    name = db.Column(db.String(64))
    code_name = db.Column(db.String(64), index=True)

    groups = db.relationship('Group', secondary='auth_groups_permissions')

    class Meta:
        permissions = (
            ('can_get_permissions', 'Can get permissions'),
        )

    def dict(self):
        return {'name': self.name, 'code_name': self.code_name}


class Group(db.Model, AbstractModelWithPermission):
    __tablename__ = "auth_groups"
    id = db.Column(db.Integer, primary_key=True, index=True)
    name = db.Column(db.String(64))

    permissions = db.relationship('Permission', secondary='auth_groups_permissions')

    class Meta:
        permissions = (
            ('can_get_groups', 'Can get groups'),
        )

    def add_permission(self, permission):
        perm = Permission.query.filter_by(code_name=permission).first()
        self.permissions.append(perm)

    def can(self, permission):
        if permission in [perm.code_name for perm in self.permissions]:
            return True
        return False

    def dict(self):
        return {
            'id': self.id, 'name': self.name,
            'permissions': [p.dict() for p in self.permissions]
        }


class GroupPermission(db.Model):
    __tablename__ = "auth_groups_permissions"
    id = db.Column(db.Integer, primary_key=True)
    group_id = db.Column(db.Integer, db.ForeignKey('auth_groups.id'))
    permission_id = db.Column(db.Integer, db.ForeignKey('auth_permissions.id'))


user_group_table = db.Table(
    "auth_users_groups",
    db.Column("user_id", db.Integer, db.ForeignKey("auth_users.id")),
    db.Column("group_id", db.Integer, db.ForeignKey("auth_groups.id")),
)


class User(db.Model, AbstractModelWithPermission, UserMixin):
    __tablename__ = 'auth_users'
    id = db.Column(db.Integer, primary_key=True, index=True)

    first_name = db.Column(db.String(64), nullable=False, default="")
    last_name = db.Column(db.String(64), nullable=False, default="")
    username = db.Column(db.String(64), index=True, unique=True)

    _email = db.Column(db.String(128), index=True, unique=True, nullable=False)
    password_hash = db.Column(db.String(128))

    groups = db.relationship("Group", secondary=user_group_table)

    is_admin = db.Column(db.Boolean, default=False)
    is_staff = db.Column(db.Boolean, default=False)

    confirmed = db.Column(db.Boolean, default=False)
    deleted = db.Column(db.Boolean, default=False)

    tokens = db.Column(db.JSON, default=[])

    class Meta:
        permissions = (
            ('can_get_users', 'Can get users'),
        )

    @staticmethod
    def _bootstrap(count):
        from mimesis import Person
        person = Person('en')

        for _ in range(count):
            u = User(
                email=person.email(),
                confirmed=True,
                first_name=person.name(),
                last_name=person.surname(),
            )

            db.session.add(u)
            try:
                db.session.commit()
            except IntegrityError:
                db.session.rollback()

    @property
    def email(self):
        return self._email

    @email.setter
    def email(self, email):
        self._email = email.lower()

    @property
    def password(self):
        return self.password_hash

    @password.setter
    def password(self, _password):
        self.password_hash = bcrypt.hashpw(_password.encode('utf8'),
                                           bcrypt.gensalt()).decode()

    def check_password(self, _password):
        return bcrypt.checkpw(_password.encode('utf8'),
                              self.password.encode('utf8'))

    def generate_token(self, token_type):
        token = random_string(length=6)
        if not self.tokens:
            self.tokens = []
        self.tokens.append({"token_type": token_type, "expire_at": time.time() + 3600, "token": token})
        if len(self.tokens) > 6:
            self.tokens.pop(0)
        flag_modified(self, "tokens")
        db.session.add(self)
        db.session.commit()
        return token

    def confirm_token(self, token, token_type):
        for t in self.tokens:
            if t['token'] == token and t['token_type'] == token_type and t['expire_at'] > time.time():
                print(token)
                self.tokens.remove(t)
                flag_modified(self, "tokens")
                db.session.add(self)
                db.session.commit()
                return True
        return False

    def set_admin(self):
        self.is_admin = True
        db.session.commit()

    def set_staff(self):
        self.is_staff = True
        db.session.commit()

    def get_confirmed(self):
        if self.confirmed:
            return gettext("بله")
        return gettext("خیر")

    def get_first_name(self):
        if self.first_name:
            return self.first_name
        return ""

    def get_last_name(self):
        if self.last_name:
            return self.last_name
        return ""

    def generate_registration_email(self):
        token = self.generate_token(CONFIRM_EMAIL)
        return {
            "subject": gettext("Verify your email account"),
            "html": render_template(TEMPLATES['confirm_email']['html'], token=token),
            "text": render_template(TEMPLATES['confirm_email']['txt'], token=token),
            "from": current_app.config['REGISTERATION_MAIL_SENDER'],
            "to": [
                self.email,
            ]
        }

    def confirm_email_with_token(self, token):
        if self.confirm_token(token, CONFIRM_EMAIL):
            self.confirmed = True
            db.session.add(self)
            db.session.commit()
            return True
        return False

    def can(self, permission):
        if self.is_admin:
            return True
        elif self.group:
            return self.group.can(permission)
        return False

    def dict(self):
        obj = {
            'id': self.id, 'first_name': self.first_name, 'last_name': self.last_name,
            'email': self.email,  'is_confirmed': self.confirmed,
        }
        return obj

    def access_dict(self):
        obj = {}
        if self.group:
            obj['group'] = self.group.dict()
        if self.is_admin:
            obj['is_admin'] = True
        if self.is_staff:
            obj['is_staff'] = True
        return obj

    def __repr__(self):
        return "<User email='%s' id=%d>" % (self.email, self.id)


class RevokedToken(db.Model):
    __tablename__ = 'auth_revoked_tokens'
    id = db.Column(db.Integer, primary_key=True)
    jti = db.Column(db.String(120))

    def add(self):
        db.session.add(self)
        db.session.commit()

    @classmethod
    def is_jti_blacklisted(cls, jti):
        query = cls.query.filter_by(jti=jti).first()
        return bool(query)


@jwt.user_loader_callback_loader
def identify(payload):
    return User.query.filter(User._email == payload).scalar()


@jwt.token_in_blacklist_loader
def check_if_token_in_blacklist(decrypted_token):
    jti = decrypted_token['jti']
    return RevokedToken.is_jti_blacklisted(jti)


@jwt.revoked_token_loader
def revoked_token_loader():
    return jsonify({'message': 'Token has been revoked'}), 401


@jwt.expired_token_loader
def expired_token_loader():
    return jsonify({"message": "Token has expired"}), 401


@jwt.invalid_token_loader
def invalid_token_loader(reason):
    return jsonify({"message": reason}), 422


@jwt.needs_fresh_token_loader
def needs_fresh_token_loader():
    return jsonify({"message": "Fresh token required"}), 401


@login_manager.user_loader
def load_user(user_id):
    return User.query.filter_by(id=user_id).first()



from flask import request, jsonify

__all__ = ('search')

models = []


def search():
    search_string = request.args.get('q')
    if not search_string:
        return jsonify({"message": "Search string must be provided"}), 422

    page = request.args.get('page', 1, type=int)
    per_page = request.args.get('per_page', 8, type=int)
    response = {}
    for model in models:
        key = "%ss" % model.__name__.lower()
        response[key] = [obj.dict() for obj in model.query.search(search_string).paginate(
            page=page, per_page=per_page, error_out=False
        ).items]
    return jsonify(response)
